#include "CGenericService.h"


SERVICE_STATUS_HANDLE CGenericServicePrivate::hservice = 0;
HANDLE CGenericServicePrivate::hthread                 = 0;
HANDLE CGenericServicePrivate::hevent                  = 0;
bool CGenericServicePrivate::statuspause               = false;
bool CGenericServicePrivate::statusrun                 = false;
int CGenericServicePrivate::threadrun                  = 0;
CGenericService* pService                              = 0;



CGenericService::CGenericService()
{
	if (pService)
	{
		/**
		 * @todo stop the application!
		 */
	}
	m_pSvc = new CGenericServicePrivate(this);
}



CGenericService::~CGenericService()
{
}



bool CGenericService::parseArguments(int argc, char** argv)
{
	if (argc > 1) 
	{
		if (!strcmp(argv[1], "-install")) 
		{
			return m_pSvc->installService();
		}
		else if (!strcmp(argv[1], "-remove")) 
		{
			return m_pSvc->removeService();
		}
		else
		{
			return m_pSvc->helpService();
		}
	}
	else
	{
		return m_pSvc->runService();
	}
} // !parseArguments()



bool CGenericService::aboutToTerminate()
{
	return m_pSvc->aboutToTerminate();
} // !aboutToTerminate()



void CGenericService::requestPause()
{
} // !requestPause()



void CGenericService::requestResume()
{
} // !requestResume()



std::string CGenericService::getServiceLabel() const
{
	return std::string("DefaultServiceLabel");
} // !getServiceLabel()



std::string CGenericService::getServiceName() const
{
	return std::string("DefaultServiceName");
} // !getServiceName()



void CGenericService::reportEvent(const std::string& msg, EventTypes type)
{
	FILE* f = fopen("e:\\test.txt", "a+");
	if (f)
	{
		fprintf(f, "%s\n", msg.c_str());
		fflush(f);
		fclose(f);
	}
} // !reportEvent()











CGenericServicePrivate::CGenericServicePrivate(CGenericService* parent)
{
	pService    = parent;
	hservice    = 0;
	hthread     = 0;
	hevent      = 0;
	statuspause = false;
	statusrun   = false;
	threadrun   = 1;

	m_nArgs  = 0;
	m_rgArgs = 0;
}

CGenericServicePrivate::~CGenericServicePrivate()
{
}



bool WINAPI CGenericServicePrivate::serviceMain(int argc, char** argv)
{
	char svcname[255];
	if (pService)
	{
		strcpy(svcname, pService->getServiceName().c_str());
	}
	else
	{
		strcpy(svcname, "defaultname");
	}
	hservice = ::RegisterServiceCtrlHandlerA(svcname,
		                                     (LPHANDLER_FUNCTION)CGenericServicePrivate::handler);
	if (!hservice)
	{
		return serviceExit();
	}

	ReportServiceStatus(SERVICE_START_PENDING,
						NO_ERROR, 0, 1, 3000);

	hevent = ::CreateEvent(0, TRUE, FALSE, 0);
	if (hevent == 0)
	{
		return serviceExit();
	}

	ReportServiceStatus(SERVICE_START_PENDING,
						NO_ERROR, 0, 2, 3000);

	BOOL ok = initService();
	if (!ok)
	{
		return serviceExit();
	}

	ReportServiceStatus(SERVICE_RUNNING);

	::WaitForSingleObject(hevent, INFINITE);
	return serviceExit(0);
}



void WINAPI CGenericServicePrivate::handler(DWORD code)
{
	DWORD status;

	switch (code)
	{
		case SERVICE_CONTROL_STOP:
			status = SERVICE_STOP_PENDING;
			stopService();
			while (threadrun != 0)
			{
				ReportServiceStatus(SERVICE_STOP_PENDING,
									NO_ERROR, 0, 1, 1000); 
			::Sleep(500);
			}
			return;

		case SERVICE_CONTROL_PAUSE:
			if (statusrun && !statuspause) 
			{
				ReportServiceStatus(SERVICE_PAUSE_PENDING,
									NO_ERROR, 0, 1, 1000);
				if (!pauseService())
				{
					/**
					 * @todo react!
					 */
				}
				status = SERVICE_PAUSED;
			}
			break;

		case SERVICE_CONTROL_CONTINUE:
			if (statusrun && statuspause)
			{
				ReportServiceStatus(SERVICE_CONTINUE_PENDING,
									NO_ERROR, 0, 1, 1000);
				if (!resumeService())
				{
					/** 
					 * @todo react!
					 */
				}
				status = SERVICE_RUNNING;
			}
			break;

		case SERVICE_CONTROL_INTERROGATE:
			break;

		case SERVICE_CONTROL_SHUTDOWN:
			return;

		default:
			break;
	}
	ReportServiceStatus(status);
} // !handler()



BOOL CGenericServicePrivate::ReportServiceStatus(DWORD dwCurrentState,
												 DWORD dwWin32ExitCode,
												 DWORD dwServiceSpecificExitCode,
												 DWORD dwCheckPoint,
												 DWORD dwWaitHint)
{
	SERVICE_STATUS status;
	status.dwServiceType  = SERVICE_WIN32_OWN_PROCESS;
	status.dwCurrentState = dwCurrentState;

	if (dwCurrentState == SERVICE_START_PENDING) 
	{
		status.dwControlsAccepted = 0;
	} 
	else
	{
		status.dwControlsAccepted = SERVICE_ACCEPT_STOP |
									SERVICE_ACCEPT_PAUSE_CONTINUE |
									SERVICE_ACCEPT_SHUTDOWN;
	}
	if (dwServiceSpecificExitCode == 0) 
	{
		status.dwWin32ExitCode = dwWin32ExitCode;
	} else 
	{
		status.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
	}
	status.dwServiceSpecificExitCode = dwServiceSpecificExitCode;

	status.dwCheckPoint = dwCheckPoint;
	status.dwWaitHint   = dwWaitHint;

	BOOL ok = ::SetServiceStatus(hservice, &status);
	if (!ok)
	{
		stopService();
	}
	return ok;
} // !ReportServiceStatus()



DWORD CGenericServicePrivate::ServiceThread(LPDWORD param)
{
	threadrun=1;
	pService->run(threadrun, 0);
	threadrun=0;
	SetEvent(hevent);
	return 0;
}



void CGenericServicePrivate::ErrorHandler(char* s)
{
	::ExitProcess(::GetLastError());
}



bool CGenericServicePrivate::initService()
{
	DWORD id;
	hthread = ::CreateThread(0, 0,
							(LPTHREAD_START_ROUTINE)CGenericServicePrivate::ServiceThread,
							 0, 0, &id);

	if (!hthread) 
	{
		return FALSE;
	}
	else 
	{
		statusrun = TRUE;
		return TRUE;
	}
} // !initService()



bool CGenericServicePrivate::pauseService()
{
	if (pService)
	{
		 pService->requestPause();
	}

	statuspause = TRUE;
	::SuspendThread(hthread);

	return true;
}



bool CGenericServicePrivate::resumeService()
{
	if (pService)
	{
		pService->requestResume();
	}
	statuspause = FALSE;
	
	return true;
} // !resumeService()



void CGenericServicePrivate::stopService()
{
	statusrun = FALSE;
	threadrun = 2;
} // !stopService()



bool CGenericServicePrivate::helpService()
{
	return true;
} // !helpService()



bool CGenericServicePrivate::installService()
{
	if (pService)
	{
		pService->reportEvent("Installing service", CGenericService::Information);
	}
	char filename[_MAX_PATH];
	::GetModuleFileNameA(NULL, filename, sizeof(filename));

	SC_HANDLE scm = ::OpenSCManager(0, 0,
		SC_MANAGER_CREATE_SERVICE);
	if (!scm)
	{
		if (pService)
		{
			pService->reportEvent("Error opening SCManager", CGenericService::Error);
		}
		CGenericServicePrivate::ErrorHandler("OpenScManager()");
	}

	char svcname[255];
	char svclabel[255];
	if (pService)
	{
		strcpy(svcname, pService->getServiceName().c_str());
		strcpy(svclabel, pService->getServiceLabel().c_str());
	}
	else
	{
		strcpy(svcname, "defaultname");
		strcpy(svclabel, "defaultlabel");
	}

	SC_HANDLE service = ::CreateServiceA(scm,
									 	 svcname,
										 svclabel,
										 SERVICE_ALL_ACCESS,
										 SERVICE_WIN32_OWN_PROCESS,
										 SERVICE_DEMAND_START,
										 SERVICE_ERROR_NORMAL,
										 filename,
										 0, 0, 0, 0, 0);
	if (!service) 
	{
		if (pService)
		{
			pService->reportEvent("Error starting service.", CGenericService::Error);
		}
		ErrorHandler("CreateService()");
	} 
	else 
	{
		if (pService)
		{
			pService->reportEvent("Service created.", CGenericService::Information);
		}
	}
	::CloseServiceHandle(service);
	::CloseServiceHandle(scm);
	return 0;
}



bool CGenericServicePrivate::removeService()
{
	SC_HANDLE scm = ::OpenSCManager(0, 0,
									SC_MANAGER_CREATE_SERVICE);
	if (!scm)
	{
		if (pService)
		{
			pService->reportEvent("Error opening SCManager", CGenericService::Error);
		}
		ErrorHandler("OpenScManager()");
	}

	char svcname[255];
	if (pService)
	{
		strcpy(svcname, pService->getServiceName().c_str());
	}
	else
	{
		strcpy(svcname, "defaultname");
	}
	SC_HANDLE service = ::OpenServiceA(scm,
									   svcname, SERVICE_ALL_ACCESS | DELETE);
	if (!service)
	{
		if (pService)
		{
			pService->reportEvent("Error opening service.", CGenericService::Error);
		}
		ErrorHandler("OpenService()");
	}

	SERVICE_STATUS status;
	BOOL ok = ::QueryServiceStatus(service, &status);
	if (!ok)
	{
		if (pService)
		{
			pService->reportEvent("Error querying service status", CGenericService::Error);
		}
		ErrorHandler("QueryServiceStatus()");
	}

	if (status.dwCurrentState!=SERVICE_STOPPED) 
	{
		if (pService)
		{
			pService->reportEvent("Service is being stopped.", CGenericService::Information);
		}
		ok = ::ControlService(service, SERVICE_CONTROL_STOP, &status);
		if (!ok) 
		{
			if (pService)
			{
				pService->reportEvent("Error controlling service (stop action)",
					                  CGenericService::Error);
			}
			ErrorHandler("ControlService()");
		}
		::Sleep(1000);
	}

	ok = ::DeleteService(service);
	if (ok) 
	{
		if (pService)
		{
			pService->reportEvent("Service deleted.", CGenericService::Information);
		}
	}
	else 
	{
		if (pService)
		{
			pService->reportEvent("Error deleting service.", CGenericService::Error);
		}
		ErrorHandler("DeleteService()");
	}
	::CloseServiceHandle(service);
	::CloseServiceHandle(scm);
	return 0;
} // !removeService()



bool CGenericServicePrivate::serviceExit(int error)
{
	if (hevent) 
	{
		::CloseHandle(hevent);
	}
	if (hservice) 
	{
		ReportServiceStatus(SERVICE_STOPPED,
							(error == -1) ? ::GetLastError() : error,
							0, 0, 0);
	}
	if (hthread) 
	{
		CloseHandle(hthread);
	}
	return 0;
}




bool CGenericServicePrivate::aboutToTerminate()
{
	return (threadrun>1);
} // !aboutToTerminate()



bool CGenericServicePrivate::runService()
{
	SERVICE_TABLE_ENTRYA servicelist[2];

	char svcname[255];
	if (!pService)
	{
		strcpy(svcname, pService->getServiceName().c_str());
	}
	else
	{
		strcpy(svcname, "defaultname");
	}

	servicelist[0].lpServiceName = svcname;
	servicelist[0].lpServiceProc = (LPSERVICE_MAIN_FUNCTIONA)CGenericServicePrivate::serviceMain;
	servicelist[1].lpServiceName = 0;
	servicelist[1].lpServiceProc = 0;

	BOOL ok = ::StartServiceCtrlDispatcherA(servicelist);
	if (!ok)
	{
		if (pService)
		{
			pService->reportEvent("Error calling ServiceCtrlDispatcher (start action)",
								  CGenericService::Error);
		}
		ErrorHandler("StartServiceCtrlDispatcher()");
	}
	return 0;
}