#ifndef _CGENERICSERVICE_H_
#define _CGENERICSERVICE_H_



#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <conio.h>
#include <windows.h>


class CGenericServicePrivate;




/**
 * @class   CGenericService
 * @author  Stephan Stapel, <stephan.stapel@web.de>
 * @date    2004-01-28
 *
 * Class for a generic service implementation
 */
class CGenericService
{
public:
	typedef enum {
		Success,
		Error,
		Warning,
		Information
	} EventTypes;
public:
	/**
	 * Standard constructor
	 */
	CGenericService();


	/**
	 * Standard destructor
	 */
	virtual ~CGenericService();


	/**
	 * Run function of the service. Needs to be implemented in
	 * inherited classes.
	 *
	 * Be sure to check aboutToTerminate() periodically in order
	 * to determine whether to continue the loop!
	 */
	virtual int run(int argc, char** argv) = 0;
	virtual void requestPause();
	virtual void requestResume();
	virtual std::string getServiceLabel() const;
	virtual std::string getServiceName() const;

	virtual void reportEvent(const std::string& msg, EventTypes type);

	bool parseArguments(int argc, char** argv);

	bool aboutToTerminate();

private:
	CGenericServicePrivate* m_pSvc;
};






class CGenericServicePrivate
{
	friend class CGenericService;
public:
	CGenericServicePrivate(CGenericService* parent = 0);
	virtual ~CGenericServicePrivate();

protected:
	bool aboutToTerminate();

private:
	static bool WINAPI serviceMain(int argc, char** argv);
    static void WINAPI handler(DWORD code);
	static BOOL ReportServiceStatus(DWORD dwCurrentState,
									DWORD dwWin32ExitCode = NO_ERROR,
									DWORD dwServiceSpecificExitCode = 0,
									DWORD dwCheckPoint = 0,
									DWORD dwWaitHint = 0);
	static DWORD WINAPI ServiceThread(LPDWORD param);
	static void ErrorHandler(char* s);

	static bool runService();
	static bool pauseService();
	static bool resumeService();
	bool startService();
	static void stopService();
	bool helpService();

	static bool initService();
	bool installService();
	bool removeService();

	static bool serviceExit(int error = -1);

private:
	static SERVICE_STATUS_HANDLE hservice;
	static HANDLE hthread;
	static HANDLE hevent;
	static bool statuspause;
	static bool statusrun;
	static int threadrun;

	int m_nArgs;
	char** m_rgArgs;
};



#endif // !_CGENERICSERVICE_H_